from glob import glob
from typing import Union, List
from multiprocessing import Pool
import os

import numpy as np
import tensorflow as tf

from .dataloader import DataLoader
from .features.featureloader import FeatureLoader
from ..utils import bytestring_feature, print_one_line

AUTOTUNE = tf.data.experimental.AUTOTUNE
TFRECORD_SHARDS = 16


def to_tfrecord(path: bytes, audio: bytes, emotion: int) -> tf.train.Example:
    feature = {
        "path": bytestring_feature([path]),
        "audio": bytestring_feature([audio]),
        "emotion": bytestring_feature([emotion])
    }
    return tf.train.Example(features=tf.train.Features(feature=feature))


def write_tfrecord_file(splitted_csv: List[np.ndarray]):
    shard_path, data = splitted_csv
    with tf.io.TFRecordWriter(shard_path, options='ZLIB') as out:
        for audio_path, emotion in data:
            audio = open(audio_path, 'rb').read()
            example = to_tfrecord(
                path=bytes(audio_path, 'utf-8'),
                audio=audio,
                emotion=emotion
            )
            out.write(example.SerializeToString())
            print_one_line("Processed:", audio_path)
    print(f"\nCreated {shard_path}")


class TFRecordDataset(DataLoader):
    def __init__(self,
                 csv_paths: Union[List[str], str],
                 tfrecords_dir: str,
                 stage: str,  # name of tfrecord (EX: `train_1.tfrecord` for 1st shard of train set tfrecord)
                 feature_loader: FeatureLoader,
                 shuffle: bool = True):
        super().__init__(csv_paths=csv_paths,
                         feature_loader=feature_loader,
                         shuffle=shuffle)
        self.stage = stage
        self.tfrecords_dir = tfrecords_dir
        if not os.path.exists(self.tfrecords_dir):
            os.makedirs(self.tfrecords_dir)

    def create_tfrecords(self):
        if glob(os.path.join(self.tfrecords_dir, f'{self.stage}*.tfrecord')):
            print("TFRecord files exists.")
            return True

        print(f"Creating {self.stage}.tfrecord...")

        csv = self.read_csv()
        if len(csv) == 0:
            return False

        def get_shard_path(shard_id):
            return os.path.join(self.tfrecords_dir, f'{self.stage}_{shard_id}.tfrecord')

        shards = [get_shard_path(i+1) for i in range(TFRECORD_SHARDS)]
        splitted_entries = np.array_split(csv, TFRECORD_SHARDS)

        with Pool(TFRECORD_SHARDS) as pool:
            pool.map(write_tfrecord_file, zip(shards, splitted_entries))

        return True

    @tf.function
    def parse(self, record):
        feature_description = {
            "path": tf.io.FixedLenFeature([], tf.string),
            "audio": tf.io.FixedLenFeature([], tf.string),
            "emotion": tf.io.FixedLenFeature([], tf.string)
        }
        example = tf.io.parse_single_example(record, feature_description)

        return tf.numpy_function(
            self.preprocess,
            inp=[example["audio"], example["emotion"]],
            Tout=[tf.float32, tf.int64]
        )

    def get_dataset(self, batch_size: int) -> Union[tf.data.Dataset, None]:
        if not self.create_tfrecords:
            return None

        pattern = os.path.join(self.tfrecords_dir, f"{self.stage}*.tfrecord")
        files_ds = tf.data.Dataset.list_files(pattern)
        ignore_order = tf.data.Options()
        ignore_order.experimental_deterministic = False
        files_ds = files_ds.with_options(ignore_order)
        dataset = tf.data.TFRecordDataset(files_ds, compression_type='ZLIB', num_parallel_reads=AUTOTUNE)

        return self.process_dataset(dataset, batch_size)



from abc import ABC
from typing import Dict, Any

import tensorflow as tf
import tensorflow_io as tfio
from tensorflow.python.eager import backprop
from tensorflow.python.keras.engine import data_adapter
from tensorflow.keras.models import Model


def apply_specaugment(
        x: tf.Tensor,
        T: float = 30,
        F: float = 5,
        n_T: int = 2,
        n_F: int = 1,
        mask_value: float = 0.) -> tf.Tensor:
    for _ in range(n_F):
        x = tfio.experimental.audio.freq_mask(x, param=F)
    for _ in range(n_T):
        x = tfio.experimental.audio.time_mask(x, param=T)
    if mask_value:
        x = tf.where(tf.equal(x, 0), tf.ones_like(x) * mask_value, x)
    return x


class BaseModel(Model, ABC):

    def __init__(self, config: Dict[str, Any], **kwargs):
        super().__init__(**kwargs)
        if config is None:
            config = {}
        self.specaugment_params = config.get('spec_augment', {})

    def train_step(self, data):
        data = data_adapter.expand_1d(data)
        x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
        if len(self.specaugment_params) > 0:
            T = self.specaugment_params.get('T', 30)
            nT = self.specaugment_params.get('nT', 2)
            F = self.specaugment_params.get('F', 5)
            nF = self.specaugment_params.get('nF', 1)
            specaugment_fn = lambda sample: apply_specaugment(sample, T=T, n_T=nT, F=F, n_F=nF)
            x = tf.map_fn(specaugment_fn, x)

        with backprop.GradientTape() as tape:
            y_pred = self(x, training=True)
            loss = self.compiled_loss(
                y, y_pred, sample_weight, regularization_losses=self.losses)
        self.optimizer.minimize(loss, self.trainable_variables, tape=tape)
        self.compiled_metrics.update_state(y, y_pred, sample_weight)
        return {m.name: m.result() for m in self.metrics}



from typing import List, Dict, Any

import tensorflow as tf

from .preprocessing import dbscale, pre_emphasis, normalize_feature


class FeatureLoader:
    def __init__(self, config: Dict[str, object]):
        self.sample_rate = config.get('sample_rate', 16000)
        self.max_length = int(self.sample_rate * (config.get('max_length', None) / 1000))  # in ms

        self.frame_length = int(self.sample_rate * (config.get("frame_length", 25) / 1000))
        self.frame_shift = int(self.sample_rate * (config.get("frame_shift", 10) / 1000))

        self.num_mel_bins = config.get('num_mel_bins', 80)
        self.feature_type = config.get('feature_type', 'fbank')
        self.preemphasis = config.get('preemphasis', 0.97)

        self.vtlp = config.get('vtlp', {})
        if len(self.vtlp) > 0:
            self.alpha_range = tf.range(
                self.vtlp.get("min_alpha", 0.9),
                self.vtlp.get("max_alpha", 1.1),
                self.vtlp.get("alpha_step", 0.1),
            )
        else:
            self.alpha_range = tf.ones([1])
        self.scale_normalize = config.get('scale_normalize', True)

    @property
    def shape(self) -> List[Any]:
        if self.feature_type == 'wave':
            return [self.max_length]
        elif self.feature_type == 'fbank':
            frame_shift = self.frame_shift * 1000. / self.sample_rate
            time_dim = int((self.max_length / self.sample_rate) * (1000. / frame_shift)) if self.max_length else None
            return [time_dim, self.num_mel_bins]
        else:
            frame_shift = self.frame_shift * 1000. / self.sample_rate
            time_dim = int((self.max_length / self.sample_rate) * (1000. / frame_shift)) if self.max_length else None
            return [time_dim, None]

    def sample_alpha(self):
        idxs = tf.range(tf.shape(self.alpha_range)[0])
        ridxs = tf.random.shuffle(idxs)[0]
        sampled_alpha = tf.gather(self.alpha_range, ridxs)
        return sampled_alpha

    def stft(self, waveform: tf.Tensor) -> tf.Tensor:
        spectrogram = tf.abs(
            tf.signal.stft(
                waveform,
                frame_length=self.frame_length,
                frame_step=self.frame_shift,
                window_fn=tf.signal.hann_window,
                pad_end=True
            )
        )
        return spectrogram
        # return apply_vtlp(spectrogram, alpha=self.sample_alpha())

    def make_spectrogram(self, waveform: tf.Tensor) -> tf.Tensor:
        spectrogram = self.stft(waveform)
        return dbscale(spectrogram)

    def make_fbank(self, waveform: tf.Tensor, fmin: float = 0., fmax: float = 8000.) -> tf.Tensor:
        spectrogram = self.stft(waveform)
        mel_fbank = tf.signal.linear_to_mel_weight_matrix(
            num_mel_bins=self.num_mel_bins,
            num_spectrogram_bins=spectrogram.shape[-1],
            sample_rate=self.sample_rate,
            lower_edge_hertz=fmin, upper_edge_hertz=fmax
        )
        fbank = tf.tensordot(spectrogram, mel_fbank, 1)
        return dbscale(fbank)

    def make_mfcc(self, waveform: tf.Tensor) -> tf.Tensor:
        log_mel_spectrogram = self.make_fbank(waveform)
        return tf.signal.mfccs_from_log_mel_spectrograms(log_mel_spectrogram)

    def extract(self, waveform: tf.Tensor) -> tf.Tensor:
        waveform = waveform[:self.max_length] if self.max_length else waveform
        waveform = pre_emphasis(waveform, coeff=self.preemphasis)

        if self.feature_type == "wave":
            return waveform
        elif self.feature_type == "spectrogram":
            features = self.make_spectrogram(waveform)
        elif self.feature_type == "fbank":
            features = self.make_fbank(waveform)
        elif self.feature_type == "mfcc":
            features = self.make_mfcc(waveform)
        else:
            raise KeyError(f"Unsupported feature type `{self.feature_type}`. Only `spectrogram`, `fbank`, and `mfcc` "
                           f"available.")

        return normalize_feature(features, scaled=self.scale_normalize)



from typing import Union, List, Tuple

import tensorflow as tf

from .metrics import weighted_accuracy, unweighted_accuracy, compute_confusion_matrix
from ..datasets.features.preprocessing import chop_feature
from ..models.base_model import BaseModel


def evaluate_slice_model(
        model: BaseModel,
        X_test: Union[List[tf.Tensor], tf.Tensor],
        y_test: tf.Tensor,
        mode: str = 'chunk', **kwargs) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    n_samples = len(X_test)
    y_pred = tf.TensorArray(tf.int64, size=0, dynamic_size=True)
    for i in tf.range(n_samples):
        x_test = X_test[i]
        if mode == "chunk":
            y_pred = y_pred.write(y_pred.size(), evaluate_chunk(model, x_test), **kwargs)
        else:
            y_pred = y_pred.write(y_pred.size(), evaluate_full(model, x_test))
    y_pred = y_pred.stack()
    wa = weighted_accuracy(y_test, y_pred)
    ua = unweighted_accuracy(y_test, y_pred)
    cm = compute_confusion_matrix(y_test, y_pred)
    return wa, ua, cm


def evaluate_testloader(
        model: BaseModel,
        test_dataset: tf.data.Dataset,
        n_steps: int) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    test_iter = iter(test_dataset)
    y_pred, y_true = [], []
    for i in range(n_steps):
        x, y = next(test_iter)
        y_hat = tf.argmax(model(x), axis=-1)
        y_true.append(y)
        y_pred.append(y_hat)
    y_true = tf.concat(y_true, axis=0)
    y_pred = tf.concat(y_pred, axis=0)
    wa = weighted_accuracy(y_true, y_pred)
    ua = unweighted_accuracy(y_true, y_pred)
    cm = compute_confusion_matrix(y_true, y_pred)
    return wa, ua, cm


def evaluate_chunk(model: BaseModel, x_test: tf.Tensor, n_frames: int = 300) -> tf.Tensor:
    chunk_test = chop_feature(x_test, n_frames=n_frames)
    y_pred = tf.argmax(tf.reduce_mean(model(chunk_test, training=False), axis=0), axis=-1)
    return y_pred


def evaluate_full(model: BaseModel, x_test: tf.Tensor) -> tf.Tensor:
    return tf.argmax(model(tf.expand_dims(x_test, 0), training=False))



from typing import Union, List, Tuple, Callable, Optional, Any

import tensorflow as tf

from .dataloader import DataLoader, EMOTIONS
from .features.featureloader import FeatureLoader
from .features.padding import pad_dup
from .features.preprocessing import load_waveform, chop_feature


class SliceDataLoader(DataLoader):
    def __init__(self,
                 csv_paths: Union[List[str], str],
                 feature_loader: FeatureLoader,
                 n_frames: int = 300,
                 thresh: int = 50,
                 pad_fn: Callable = pad_dup,
                 shuffle: bool = True):
        super().__init__(feature_loader=feature_loader,
                         csv_paths=csv_paths,
                         shuffle=shuffle)
        self.n_frames = n_frames
        self.thresh = thresh
        self.pad_fn = pad_fn

    def preprocess(self, data):
        audio_path, emotion = data[0], data[1]
        with tf.device('/CPU:0'):
            waveform = load_waveform(audio_path)

            features = self.feature_loader.extract(waveform)
            chunks = chop_feature(features, n_frames=self.n_frames, thresh=self.thresh, pad_fn=self.pad_fn)
            label = tf.argmax(tf.stack([tf.equal(emotion, e) for e in EMOTIONS], axis=-1))
            return chunks, label

    def get_dataset(self, **kwargs) -> Optional[Tuple[Union[Union[tf.Tensor, List[tf.Tensor]], Any], Any]]:
        data = self.read_csv()
        if len(data) == 0:
            return None
        y = tf.TensorArray(tf.int64, size=0, dynamic_size=True)

        data_iterator = iter(data)
        d = next(data_iterator)
        x, label = self.preprocess(d)
        y = y.write(y.size(), label)
        for d in data_iterator:
            chunks, label = self.preprocess(d)
            x = tf.concat([x, chunks], axis=0)
            y = y.write(y.size(), label)
        return x, y.stack()


class TestSliceDataLoader(DataLoader):
    def __init__(self,
                 csv_paths: Union[List[str], str],
                 feature_loader: FeatureLoader,
                 shuffle: bool = True):
        super().__init__(feature_loader=feature_loader,
                         csv_paths=csv_paths,
                         shuffle=shuffle)

    def preprocess(self, data):
        audio_path, emotion = data[0], data[1]
        with tf.device('/CPU:0'):
            waveform = load_waveform(audio_path)

            features = self.feature_loader.extract(waveform)
            label = tf.argmax(tf.stack([tf.equal(emotion, e) for e in EMOTIONS], axis=-1))
            return features, label

    def get_dataset(self, **kwargs) -> Optional[Tuple[Union[Union[tf.Tensor, List[tf.Tensor]], Any], Any]]:
        data = self.read_csv()
        if len(data) == 0:
            return None
        X = list()
        y = tf.TensorArray(tf.int64, size=0, dynamic_size=True)

        data_iterator = iter(data)
        for d in data_iterator:
            x, label = self.preprocess(d)
            X.append(x)
            y = y.write(y.size(), label)
        return X, y.stack()



from typing import Callable, Union, List

import tensorflow as tf

from .padding import pad_dup


def load_waveform(
        audio: str,
        bit_depth: int = 16,
        dither: float = 1.) -> tf.Tensor:
    audio_binary = tf.io.read_file(audio)
    waveform, sr = tf.audio.decode_wav(audio_binary)
    # if sr != sample_rate:
    #     raise ValueError(f"Expected sampling rate of {sample_rate} but got {sr}")
    waveform = tf.squeeze(waveform, axis=-1)
    waveform = waveform + tf.random.normal(tf.shape(waveform), stddev=(dither / 2 ** bit_depth))
    return waveform


def pre_emphasis(waveform: tf.Tensor, coeff: float = 0.97) -> tf.Tensor:
    if isinstance(coeff, float) or coeff <= 0.0:
        return waveform
    return tf.concat([waveform[0], waveform[1:] - coeff * waveform[:-1]])


def normalize_feature(feature: tf.Tensor, scaled: bool = False, epsilon: float = 1e-8) -> tf.Tensor:
    mu = tf.reduce_mean(feature, axis=0)
    sum_x2 = tf.math.reduce_sum(feature ** 2, axis=0) / tf.cast(tf.shape(feature)[0], dtype=tf.float32)
    sigma = sum_x2 - (mu ** 2) + epsilon
    normalized = (feature - mu) / sigma if scaled else feature - mu
    return normalized


def dbscale(spectrogram: tf.Tensor, top_db: int = 80):
    power = tf.math.square(spectrogram)
    log_spec = 10.0 * (tf.math.log(power) / tf.math.log(10.0))
    log_spec = tf.math.maximum(log_spec, tf.math.reduce_max(log_spec) - top_db)
    return log_spec

# ----------------- VTLP ----------------------


def chop_feature(
        x: tf.Tensor,
        n_frames: int,
        thresh: int = 50,
        pad_fn: Callable = pad_dup) -> Union[tf.Tensor, List[tf.Tensor]]:
    """Chop fbank into smaller chunk. Pad function is
    optional. If not specified, this function will return
    a list of preprocessing.
    """
    time_dim = tf.shape(x)[0]
    x_chopped = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
    for i in tf.range(time_dim):
        if i % n_frames == 0 and i != 0:  # if reach n_frames, cut
            xi = x[i - n_frames:i]
            x_chopped = x_chopped.write(x_chopped.size(), xi)
    if i < time_dim and i < n_frames:  # if file length is less than n_frames, pad
        xi = pad_fn(x, max_len=n_frames)
        x_chopped = x_chopped.write(x_chopped.size(), xi)
    else:  # if file is longer than n_frame, pad remainder
        remainder = x[time_dim - time_dim % n_frames:]
        t_remainder = tf.shape(remainder)[0]
        if t_remainder > thresh:
            xi = pad_fn(remainder, max_len=n_frames)
            x_chopped = x_chopped.write(x_chopped.size(), xi)
    if pad_fn:
        return x_chopped.stack()
    return x_chopped

# ----------------- VTLP (Experimental) ----------------------


def get_scale_factors(freq_dim, sampling_rate, f_hi=4800., alpha=0.9):
    factors = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
    freqs = tf.cast(tf.linspace(0, 1, freq_dim), tf.float32)

    scale = f_hi * tf.minimum(alpha, 1.)
    f_boundary = scale / alpha
    half_sr = sampling_rate / 2

    for i in tf.range(len(freqs)):
        f = sampling_rate * freqs[i]
        if tf.math.less_equal(f, f_boundary):
            factors = factors.write(factors.size(), f * alpha)
        else:
            warp_freq = half_sr - (half_sr - scale) / (half_sr - scale / alpha) * (half_sr - f)
            factors = factors.write(factors.size(), warp_freq)
    return factors.stack()


def apply_vtlp(spec, sampling_rate: int = 16000, f_hi: int = 4800, alpha=0.9):
    time_dim, freq_dim = tf.shape(spec)[0], tf.shape(spec)[1]
    factors = get_scale_factors(freq_dim, sampling_rate, f_hi=f_hi, alpha=alpha)
    factors = factors * tf.cast(freq_dim - 1, tf.float32) / tf.reduce_max(factors)
    new_spec = tf.zeros([freq_dim, time_dim], dtype=tf.float32)

    for i in tf.range(freq_dim):
        if tf.equal(i, 0) or tf.greater_equal(i + 1, freq_dim):
            update = tf.expand_dims(spec[:, i], 0)
            new_spec = tf.tensor_scatter_nd_add(new_spec, [[i]], update)
        else:
            warp_up = factors[i] - tf.floor(factors[i])
            warp_down = 1. - warp_up
            pos = tf.cast(tf.floor(factors[i]), tf.int64)

            new_spec = tf.tensor_scatter_nd_add(new_spec, [[pos]], tf.expand_dims(warp_down * spec[:, i], 0))
            new_spec = tf.tensor_scatter_nd_add(new_spec, [[pos + 1]], tf.expand_dims(warp_up * spec[:, i], 0))
    return tf.transpose(new_spec, perm=(1, 0))



from typing import Callable, List

import tensorflow as tf


def pad_dup(x: tf.Tensor, max_len: tf.Tensor) -> tf.Tensor:
    """Pad an Arguments feature upto specified length.
    The Arguments is repeated until max_len is reached.
    """
    time_dim = tf.cast(tf.shape(x)[0], tf.float32)

    temp = tf.identity(x)
    num_repeat = tf.floor(max_len / time_dim)
    remainder = tf.cast(max_len - (num_repeat * time_dim), tf.int64)
    x_rem = x[:remainder, :]
    for _ in tf.range(num_repeat-1):
        x = tf.concat([x, temp], axis=0)
    x_pad = tf.concat([x, x_rem], axis=0)
    return x_pad


def pad_zero(x: tf.Tensor, max_len: tf.Tensor) -> tf.Tensor:
    """Pad Arguments feature up to specified length.
    The padded values are zero.
    Arguments
    """
    time_dim, freq_dim = tf.shape(x)[0], tf.shape(x)[1]
    time_dim = tf.cast(time_dim, tf.float32)
    freq_dim = tf.cast(freq_dim, tf.float32)

    zeros = tf.zeros((max_len - time_dim, freq_dim))
    x_pad = tf.concat([x, zeros], axis=0)
    return x_pad


def pad_X(X: List[tf.Tensor], pad_fn: Callable, max_len: int = None) -> tf.Tensor:
    """Pad a pack of array to a specified max_len.
    If max_len is not specified, longest preprocessing will
    be use as a max length. This function is used to
    pad the fbank features to its max_len, not chopped.
    Arguments
    """
    if not max_len:
        max_len = tf.cast(tf.maximum([tf.shape(x)[0] for x in X]), tf.float32)
    out = [pad_fn(x, max_len) for x in X]
    return tf.convert_to_tensor(out)